package ppds;

import static org.junit.Assert.assertEquals;

import org.apache.commons.lang3.tuple.Pair;
import org.junit.Test;

import ppds.Types.ErrCode;
import ppds.Types.KeyType;

public class InsertTest {
    public IdxState ixstate = null;
    public TxnState txstate = null;

    public KeyType type = KeyType.INT;
    public IXServer ixserver = new IXServer();
    private Integer PAGE_SIZE = 4096;

    @Test
    public void insertTest() throws InterruptedException {

        ErrCode ret1 = null;
        ret1 = ixserver.create(type, "primary_index", PAGE_SIZE);
        assertEquals("could not create primary index", ErrCode.SUCCESS.name(), ret1.toString());

        Pair<ErrCode, IdxState> pIdx =  ixserver.openIndex("primary_index");
        ixstate = pIdx.getRight();
        assertEquals("could not open primary index", ErrCode.SUCCESS.name(), pIdx.getLeft().toString());

        Pair<ErrCode, TxnState> pTxn = ixserver.beginTransaction();
        txstate = pTxn.getRight();
        assertEquals("could not begin transaction", ErrCode.SUCCESS.name(), pTxn.getLeft().toString());

        for (int i = 0; i < 100; i++) {
           Key<?> key = new Key<>(KeyType.INT, i);

           String value = "value ";
           if (i < 10)
               value = value.concat("0" + i);
           else
                value = value.concat(String.valueOf(i));

            ret1 = ixserver.insertRecord(ixstate, txstate, key, value);

            assertEquals("failed to insert " + key.getKeyval() + " " + value, ErrCode.SUCCESS.name(), ret1.toString());
        }

        if (ixserver.commitTransaction(txstate) != ErrCode.SUCCESS) {
            System.out.println("Could not end primary index txn");
            return;
        }
        if (ixserver.closeIndex(ixstate) != ErrCode.SUCCESS) {
            System.out.println("Could not close primary index");
        }
    }
}
